package loopbounds.ng;

import java.lang.reflect.InvocationTargetException;
import java.util.List;

/** 
 * Interface for node generators. A node generator is a class that generates
 * test-nodes that lie in NCA configuration, by some algorithm.
 * 
 * @author Rody Kersten
 *
 */
public interface NodeGenerator {
	
	/** Generate the test-nodes */
	void generate() throws NotEnoughTestNodesException, InvocationTargetException;
	
	/** Get the number of test nodes found.
	 * 
	 * @return The number of nodes
	 */
	public int getNumberOfNodes();
	
	/** Get the nodes.
	 * 
	 * @return a list of test-nodes in NCA configuration
	 */
	public List<int[]> getNodes();
}
