package loopbounds.ng;

/** Expression for when not enough test-nodes can be found that satisfy
 * both the loop condition and NCA-configuration.
 * 
 * @author Rody Kersten
 *
 */
public class NotEnoughTestNodesException extends Exception {
	static final long serialVersionUID = 220;
	
	private String condition;
	private int minval, maxval;
	
	/** Constructor.
	 * 
	 * @param condition The condition for which test-nodes cannot be found
	 * @param minval The minimal bound, above which test-nodes are searched
	 * @param maxval The maximal bound, below which test-nodes are searched
	 */
	public NotEnoughTestNodesException(String condition, int minval, int maxval) {
		this.condition = condition;
		this.minval = minval;
		this.maxval = maxval;
	}
	
	/** 
	 * Outputs a String describing for which condition test-nodes could not be found,
	 * between which bounds.
	 * 
	 * @return A String describing for which condition test-nodes could not be found,
	 * between which bounds.
	 */
	@Override
	public String toString() {
		return new String(
			"Not enough test-nodes can be found that satisfy NCA-configuration and loop condition "
			+ condition + " between " + minval + " and " + maxval + ". ");
	}
}
