package loopbounds.parsetree;

import java.util.List;
import java.util.ArrayList;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;
import org.jmlspecs.openjml.JmlTree.JmlClassDecl;

import com.sun.tools.javac.tree.JCTree;

/**
 * Class for finding the class names in a JCTree. It is intended
 * to be used through the static function 'getClasses(JCTree tree)' which
 * takes a tree node as an argument and returns a List of the classes in the tree.
 * <P>
 * The class extends JmlTreeScanner and uses the visitor pattern.
 * 
 * @author Rody Kersten
 *
 */
public class ClassScanner extends JmlTreeScanner implements IJmlVisitor {
	
	private List<JmlClassDecl> classes = new ArrayList<JmlClassDecl>();
		
	/** Gets the classes found in the JCTree.
	 * 
	 * @param tree The parse tree to search
	 * @return A list of the classes in 'tree'
	 */
	public static List<JmlClassDecl> getClasses (JCTree tree) {
		ClassScanner scanner = new ClassScanner();
		scanner.scan(tree);
		return scanner.classes;
	}
	
	@Override
	public void visitJmlClassDecl (JmlClassDecl tree) {
		classes.add(tree);
		// No call of super.visitJmlClassDecl(tree), tree scanning is stopped.
	}
}
