package loopbounds.parsetree;

import java.util.ArrayList;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTree.JmlClassDecl;
import org.jmlspecs.openjml.JmlTree.JmlVariableDecl;
import org.jmlspecs.openjml.JmlTreeScanner;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCExpression;

/**
 * Class for finding the type of a variable in a parse tree.
 * <P>
 * The class extends JmlTreeScanner and uses the visitor pattern.
 * 
 * @author Joel Haasnoot
 *
 */
public class ClassTypeScanner extends JmlTreeScanner implements IJmlVisitor {
	
	private ArrayList<CodeClass> classes;
	private String curClass;
	private ArrayList<String> curVars;
	private ArrayList<JCExpression> curTypes;
	
	private ClassTypeScanner() {
		this.classes = new ArrayList<CodeClass>();
		this.curClass = "";
		this.curVars = new ArrayList<String>();
		this.curTypes = new ArrayList<JCExpression>();
	}
	
	/** Finish the scanning process and add the last class found to the ArrayList. */
	private void finish(){
		if (curClass != "") {
			classes.add(new CodeClass(curClass, curVars, curTypes));
		}
	}
	
	/** Scan the JCTree 'tree' for any and all class declarations.
	 * 
	 * @param tree The parse tree in which we are looking
	 * @return ArrayList of JmlClassDecl which are class declarations
	 */
	public static ArrayList<CodeClass> getTypes(JCTree tree) {
		ClassTypeScanner ts = new ClassTypeScanner();
		ts.scan(tree);
		ts.finish(); // Add the last class to the ArrayList.
		ts.classes.trimToSize();
		return ts.classes;
	}
	
	@Override
	public void visitJmlVariableDecl (JmlVariableDecl tree) {
		if (!curVars.contains(tree.name.toString())){
			curVars.add(tree.name.toString());
			curTypes.add(tree.vartype);
		}
		super.visitJmlVariableDecl(tree);
	}
	
	@Override
	public void visitJmlClassDecl (JmlClassDecl that) {
		if (curClass != "") {
			classes.add(new CodeClass(curClass, curVars, curTypes));
			curVars.clear();
			curTypes.clear();
		}
		curClass = that.name.toString();
		super.visitJmlClassDecl(that);
	}
}