package loopbounds.parsetree;

import java.util.ArrayList;

import com.sun.tools.javac.tree.JCTree.JCExpression;


/**
 * Class to store information about classes and their member variables and types
 * Used to recursively evaluate usage of fields
 * 
 * @author Joel Haasnoot
 *
 */
public class CodeClass {
	
	private String name;
	private String[] fields;
	private JCExpression[] types;
	
	
	/**
	 * @param name Name of class
	 * @param fields Array of fields
	 * @param types Array of types of the fields
	 */
	public CodeClass(String name, String[] fields, JCExpression[] types) {
		this.name = name;
		this.fields = fields;
		this.types = types;
	}
	
	/** Alias construct, converts ArrayList to array 
	 * @param name Name of class
	 * @param vars ArrayList of fields
	 * @param types ArrayList of variable types in the same order as the names array
	 */
	public CodeClass(String name, ArrayList<String> fields, ArrayList<JCExpression> types) {
		this.name = name;
		this.fields = new String[fields.size()];
		for (int i = 0; i < fields.size(); i++)
			this.fields[i] = fields.get(i);
		this.types = new JCExpression[types.size()];
		for (int i = 0; i < types.size(); i++)
			this.types[i] = types.get(i);	
	}

	/** Get the name of the class.
	 * 
	 * @return the name of the class
	 */
	public String getName() {
		return name;
	}
	
	/** Get the fields.
	 * 
	 * @return the field of this class
	 */
	public String[] getFields() {
		return fields;
	}
	
	/** Get the field types.
	 * 
	 * @return the types of the fields of this class
	 */
	public JCExpression[] getTypes() {
		return types;
	}
}
