package loopbounds.parsetree;

import java.util.ArrayList;
import java.util.List;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;
import org.jmlspecs.openjml.JmlTree.JmlDoWhileLoop;
import org.jmlspecs.openjml.JmlTree.JmlForLoop;
import org.jmlspecs.openjml.JmlTree.JmlWhileLoop;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCStatement;

/**
 * Class for finding loops in a parse tree.
 * <P>
 * The class extends JmlTreeScanner and uses the visitor pattern.
 * 
 * @author Rody Kersten
 * 
 */
public class LoopScanner extends JmlTreeScanner implements IJmlVisitor {

	private List<JCStatement> loops = new ArrayList<JCStatement>();

	/**
	 * Get a list of a the loops in the given parse tree.
	 * 
	 * @param tree
	 *            An OpenJml parse tree
	 * @return A list of the while, do-while and for loops in the given tree
	 */
	public static List<JCStatement> getLoops(JCTree tree) {
		LoopScanner scanner = new LoopScanner();
		scanner.scan(tree);
		return scanner.loops;
	}

	@Override
	public void visitJmlWhileLoop(JmlWhileLoop tree) {
		loops.add((JCStatement) tree);
		super.visitJmlWhileLoop(tree);
	}

	@Override
	public void visitJmlForLoop(JmlForLoop tree) {
		loops.add((JCStatement) tree);
		super.visitJmlForLoop(tree);
	}

	@Override
	public void visitJmlDoWhileLoop(JmlDoWhileLoop tree) {
		loops.add((JCStatement) tree);
		super.visitJmlDoWhileLoop(tree);
	}
}
