package loopbounds.parsetree;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;
import org.jmlspecs.openjml.JmlTree.JmlClassDecl;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCMethodDecl;

/**
 * Class for adding a method to a class in the parse tree.
 * <P>
 * The class extends the JmlTreeScanner class and uses the visitor pattern. When
 * the JmlClassDecl is encountered the methods are replaced.
 * No subsequent nodes are visited.
 * 
 * @author Rody Kersten
 *
 */
public class MethodAdder extends JmlTreeScanner implements IJmlVisitor {
	
	com.sun.tools.javac.util.List<com.sun.tools.javac.tree.JCTree> meths;
	int curIndex, classIndex;
	
	/** Constructor. Create a new MethodAdder.
	 * 
	 * @param meths List of the methods to add.
	 * @param classIndex The index of the class in which to add the method.
	 */
	private MethodAdder(com.sun.tools.javac.util.List<com.sun.tools.javac.tree.JCTree> meths, int classIndex) {
		this.meths = meths;
		this.classIndex = classIndex;
		this.curIndex = 0;
	}
	
	/** Add methods.
	 * 
	 * @param meths The methods to add.
	 * @param tree The parse tree in which to add the methods
	 * @param classIndex The index of the class in the parse tree, in which to add the methods.
	 */
	public static void add(com.sun.tools.javac.util.List<com.sun.tools.javac.tree.JCTree> meths, JCTree tree, int classIndex) {
		MethodAdder ma = new MethodAdder(meths, classIndex);
		ma.scan(tree);
	}
	
	/** Adds a single method.
	 * 
	 * @param m The new method
	 * @param tree The parse tree to add the method to
	 * @param classIndex The index of the class in the parse tree, in which to add the method.
	 */
	public static void addSingle(JCMethodDecl m, JCTree tree, int classIndex) {
		com.sun.tools.javac.util.List<com.sun.tools.javac.tree.JCTree> meths = com.sun.tools.javac.util.List.<com.sun.tools.javac.tree.JCTree>of(m);
		MethodAdder.add(meths, tree, classIndex);
	}
	
	@Override
	public void visitJmlClassDecl (JmlClassDecl tree) {
		if (curIndex==classIndex) {
			//add to the end
			tree.defs = tree.defs.appendList(meths);
		}
		curIndex++;
		// No call of super.visitJmlClassDecl(tree), no deeper tree scanning.
	}
}

