package loopbounds.parsetree;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;
import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCArrayAccess;
import com.sun.tools.javac.tree.JCTree.JCFieldAccess;
import com.sun.tools.javac.tree.JCTree.JCIdent;
import com.sun.tools.javac.tree.JCTree.JCMethodInvocation;
import com.sun.tools.javac.tree.JCTree.JCVariableDecl;

public class ObjectScanner extends JmlTreeScanner implements IJmlVisitor {

	boolean inApply = false;
	boolean inArg = false;
	ObjectSet usedObjects = new ObjectSet();
	ObjectSet declaredObjects = new ObjectSet();

	public void visitVarDef(JCVariableDecl tree) {
		declaredObjects.add(tree.name.toString());
		scan(tree.mods);
		scan(tree.vartype);
		scan(tree.init);
	}

	public void visitApply(JCMethodInvocation tree) {
		inApply = true;
		scan(tree.meth);
		inApply = false;
		inArg = true;
		scan(tree.args);
		inArg = false;
	}

	public void visitSelect(JCFieldAccess tree) {
		scan(tree.selected);
	}
	
	public void visitIndexed(JCArrayAccess tree) {
        scan(tree.indexed);
    }

	public void visitIdent(JCIdent tree) {
		if (inApply || inArg) {
			usedObjects.add(tree.toString());
		}
	}

	public static ObjectSet getUsedObjects(JCTree tree) {
		ObjectScanner scanner = new ObjectScanner();
		scanner.scan(tree);
		return scanner.usedObjects;
	}

	public static ObjectSet getDeclaredObjects(JCTree tree) {
		ObjectScanner scanner = new ObjectScanner();
		scanner.scan(tree);
		return scanner.declaredObjects;
	}
}
