package loopbounds.parsetree;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;
import org.jmlspecs.openjml.JmlTree;

import com.sun.tools.javac.util.Context;
import com.sun.tools.javac.util.Name;
import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCReturn;

/**
 * Class for replacing all return statements in an abstract syntax tree. The class
 * is intended to be used through the static function 'replace'. Which replaces all
 * return statements in the given tree, with with a return statement that returns
 * the given variable.
 * <P>
 * The class extends JmlTreeScanner and uses the visitor pattern.
 * 
 * @author Martijn Moraal
 *
 */
public class ReturnReplacer extends JmlTreeScanner implements IJmlVisitor {
	
	private JmlTree.Maker M;
	private Name name;
	
	private ReturnReplacer (Name name, Context context) {
		this.M = JmlTree.Maker.instance(context);
		this.name = name;
	}
	
	/** Replaces all the return statements in the given tree with a statement
	 * that returns the given variable.
	 * 
	 * @param varname The variable to be returned
	 * @param tree The tree to replace the return statements in
	 * @param context The compilation context from OpenJDK
	 */
	public static void replace (Name varname, JCTree tree, Context context) {
		ReturnReplacer replacer = new ReturnReplacer(varname, context);
		replacer.scan(tree);
	}
	
	@Override
	public void visitReturn (JCReturn tree) {
		tree.expr = M.Ident(name);
		super.visitReturn(tree);
	}
}
