package tests;

public class typeclauses extends TCBase {

    String eol = "\n";  // Just newline for Windows also
    
    @Override
    public void setUp() throws Exception {
//        noCollectDiagnostics = true;
//        jmldebug = true;
        super.setUp();
    }

    /** Tests typechecking an invariant clause - OK*/
    public void testInvariant() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ invariant b;\n}");
    }

    /** Tests typechecking an invariant clause - bad type*/
    public void testInvariant2() {
        helpTCF("A.java"," class A { int k; boolean b; Boolean bb; \n//@ invariant k;\n}",
                "/A.java:2: incompatible types\n  required: boolean\n  found:    int",15);
    }

    /** Tests typechecking an invariant clause - OK from Boolean*/
    public void testInvariant3() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ invariant bb;\n}");
    }

    /** Tests static lookup for invariant */
    public void testInvariant4() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ static invariant bb;\n}"
                ,"/TEST.java:2: non-static variable bb cannot be referenced from a static context",22
                );
    }

    /** Tests static lookup for invariant */
    public void testInvariant5() {
        helpTC(" class A { int k; boolean b; static Boolean bb; \n//@ static invariant bb;\n}");
    }

    /** Tests typechecking an constraint clause - OK*/
    public void testConstraint() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ constraint b;\n}");
    }

    /** Tests typechecking an constraint clause - bad type*/
    public void testConstraint2() {
        helpTCF("A.java"," class A { int k; boolean b; Boolean bb; \n//@ constraint k;\n}",
                "/A.java:2: incompatible types\n  required: boolean\n  found:    int",16);
    }

    /** Tests typechecking an constraint clause - OK from Boolean*/
    public void testConstraint3() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ constraint bb;\n}");
    }

    /** Tests typechecking an constraint clause - OK from Boolean*/
    public void testConstraint4() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ constraint bb for \\everything;\n}");
    }

    /** Tests typechecking an constraint clause - OK from Boolean*/
    public void testConstraint5() {
        helpTC(" class A { void m(int i) {} Boolean bb; \n//@ constraint bb for A(), m, m(int), m(Object);\n}"
                ,"/TEST.java:2: Constructors are not allowed as methods in constraint clauses",23
                ,"/TEST.java:2: method m in class A cannot be applied to given types;\n  required: int\n  found: java.lang.Object\n  reason: actual argument java.lang.Object cannot be converted to int by method invocation conversion",39
                );
    }

    /** Tests static lookup for constraint */
    public void testConstraint6() {
        helpTC(" class A { void m(int i) {} Boolean bb; \n//@ static constraint bb ;\n}"
                ,"/TEST.java:2: non-static variable bb cannot be referenced from a static context",23
                );
    }

    /** Tests static lookup for constraint */
    public void testConstraint7() {
        helpTC(" class A { void m(int i) {} static Boolean bb; \n//@ static constraint bb ;\n}");
    }

    // TODO - the testConstraintM... tests are not implemented
    public void testConstraintM() {
        helpTC(" class A { void m(int i) {} Boolean bb; \n//@ constraint bb for m;\n}");
    }

    public void testConstraintM1() {
        helpTC(" class A { void m(int i) {} Boolean bb; \n//@ constraint bb for mm;\n}");
    }

    public void testConstraintM2() {
        helpTC(" class A { void m(int i) {} Boolean bb; \n//@ constraint bb for this.m;\n}");
    }

    public void testConstraintM3() {
        helpTC(" class A { void m(int i) {} Boolean bb; \n//@ constraint bb for A.m;\n}");
    }

    public void testConstraintMA() {
        helpTC(" class A { void m(int i) {} Boolean bb; \n//@ constraint bb for m(int);\n}");
    }

    public void testConstraintMA1() {
        helpTC(" class A { void m(int i) {} Boolean bb; \n//@ constraint bb for mm(int);\n}"
                ,"/TEST.java:2: cannot find symbol\n  symbol:   method mm(int)\n  location: class A",23
                );
    }

    public void testConstraintMA2() {
        helpTC(" class A { void m(int i[],Object o) {} void m(int i) {} Boolean bb; \n//@ constraint bb for this.m(int[],Object);\n}");
    }

    public void testConstraintMA3() {
        helpTC(" class A { void m(Integer i) {} A a; Boolean bb; \n//@ constraint bb for a.m(java.lang.Integer);\n}");
    }

    public void testConstraintMA3s() {
        helpTC(" class A { static void m(Integer i) {} Boolean bb; \n//@ constraint bb for A.m(java.lang.Integer);\n}");
    }

    // FIXME - distinguish static from instance
    public void testConstraintMA3e() {
        helpTC(" class A { void m(Integer i) {} A a; Boolean bb; \n//@ constraint bb for A.m(java.lang.Integer);\n}");
    }

    // FIXME - distinguish static from instance
    public void testConstraintMA3se() {
        helpTC(" class A { static void m(Integer i) {} A a; Boolean bb; \n//@ constraint bb for a.m(java.lang.Integer);\n}");
    }

    public void testConstraintMA4() {
        helpTC(" class A extends B { Boolean bb; \n//@ constraint bb for m(java.lang.Integer);\n} class B { void m(Integer i) {} }"
                ,"/TEST.java:2: The method must be a direct member of the class containing the constraint clause",23
                );
    }

    public void testConstraintMA5() {
        helpTC(" class A { Boolean bb; \n//@ constraint bb for B.m(java.lang.Integer);\n} class B { void m(Integer i) {} }"
                ,"/TEST.java:2: The method must be a direct member of the class containing the constraint clause",23
                );
    }

    /** Tests typechecking an axiom clause - OK*/
    public void testAxiom() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ axiom b;\n}");
    }

    /** Tests typechecking an axiom clause - bad type*/
    public void testAxiom2() {
        helpTCF("A.java"," class A { int k; boolean b; Boolean bb; \n//@ axiom k;\n}",
                "/A.java:2: incompatible types\n  required: boolean\n  found:    int",11);
    }

    /** Tests typechecking an axiom clause - OK from Boolean*/
    public void testAxiom3() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ axiom bb;\n}");
    }

    /** Tests typechecking an initially clause - OK*/
    public void testInitially() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ initially b;\n}");
    }

    /** Tests typechecking an initially clause - bad type*/
    public void testInitially2() {
        helpTCF("A.java"," class A { int k; boolean b; Boolean bb; \n//@ initially k;\n}",
                "/A.java:2: incompatible types\n  required: boolean\n  found:    int",15);
    }

    /** Tests typechecking an initially clause - OK from Boolean*/
    public void testInitially3() {
        helpTC(" class A { int k; boolean b; Boolean bb; \n//@ initially bb;\n}");
    }

    /** Tests typechecking an initially clause - OK from Boolean*/
    public void testInitially4() {
        helpTCF("A.java"," class A { int k; boolean b; Boolean bb; \n//@ initially x;\n}",
                "/A.java:2: cannot find symbol\n  symbol:   variable x\n  location: class A",15);
    }

    /** Tests initially may be static */
    public void testInitially5() {
        helpTCF("A.java"," class A { int k; boolean b; Boolean bb; \n//@ static initially b;\n}"
                ,"/A.java:2: non-static variable b cannot be referenced from a static context",22
                );
    }

    public void testRepresents() {
        helpTCF("A.java","public class A {\n //@ model int i; represents i = true;\n}"
                ,"/A.java:2: incompatible types\n  required: int\n  found:    boolean",34
                );
    }

    public void testRepresents1() {
        helpTCF("A.java","public class A {\n //@ model int i; represents i <- true;\n}"
                ,"/A.java:2: warning: The left arrow is deprecated in represents clauses, use = instead",32
                ,"/A.java:2: incompatible types\n  required: int\n  found:    boolean",35
                );
    }

    public void testRepresents2() {
        helpTCF("A.java","public class A {\n //@ model int i; represents i \\such_that 0;\n}"
                ,"/A.java:2: incompatible types\n  required: boolean\n  found:    int",43
                );
    }

    public void testRepresents3() {
        helpTCF("A.java","public class A {\n //@ model int i; represents j = 0;\n}"
                ,"/A.java:2: cannot find symbol\n  symbol:   variable j\n  location: class A",30
                );
    }
    
    public void testRepresents4() {
        helpTCF("A.java","public class A {\n //@ model int i; represents i :0;\n}"
                ,"/A.java:2: A represents clause must have a = or \\such_that after the identifier",32
                );
    }
    
    public void testRepresents5() {
        helpTCF("A.java","public class A {\n //@ model int j; represents j = ;\n}"
                ,"/A.java:2: illegal start of expression",34
                );
    }
    
    public void testRepresents6() {
        helpTCF("A.java","public class A {\n //@ model int i; represents j = 0\n}"
                ,"/A.java:2: The expression is invalid or not terminated by a semicolon",35
                ,"/A.java:2: cannot find symbol"+eol+"  symbol:   variable j"+eol+"  location: class A",30
                );
    }
    
    public void testRepresents7() {
        helpTCF("A.java","public class A {\n //@ model int i; represents x = 0\n}"
                ,"/A.java:2: The expression is invalid or not terminated by a semicolon",35
                ,"/A.java:2: cannot find symbol"+eol+"  symbol:   variable x"+eol+"  location: class A",30
                );
    }
    
    public void testRepresents8() {
        helpTCF("A.java","public class A {\n //@ model int i; represents x.* = 0\n}"
                ,"/A.java:2: Expected an identifier after the dot in this context",32
                ,"/A.java:2: A represents clause must have a = or \\such_that after the identifier",37
                ,"/A.java:2: cannot find symbol"+eol+"  symbol:   variable x"+eol+"  location: class A",30
                );
    }
    
    public void testRepresents9() {
        helpTCF("A.java","public class A {\n //@ model int i; represents x[*] = 0\n}"
                ,"/A.java:2: illegal start of expression",32
                ,"/A.java:2: illegal start of expression",33
                ,"/A.java:2: The expression is invalid or not terminated by a semicolon",38
                ,"/A.java:2: cannot find symbol"+eol+"  symbol:   variable x"+eol+"  location: class A",30
                );
    }
    
    public void testRepresents10() {
        helpTCF("A.java","public class A {\n //@ model int i; represents x[3] = 0;\n}"
                ,"/A.java:2: cannot find symbol\n  symbol:   variable x\n  location: class A",30
                );
    }
    
    public void testRepresents11() {
        helpTCF("A.java","public class A {\n //@ model int i; static represents i = 0;\n}"
                ,"/A.java:2: A represents clause and its associated model field must both be static or both not be static",26
                );
    }
    
    public void testRepresents12() {
        helpTCF("A.java","public class A {\n //@ model static int i; represents i = 0;\n}"
                ,"/A.java:2: A represents clause and its associated model field must both be static or both not be static",26
                );
    }
    
    public void testRepresents13() {
        helpTCF("A.java","public class A {\n //@ ghost int i; represents i = 0;\n}"
                ,"/A.java:2: The target of a represents clause must be a model field",19
                );
    }
    
    public void testRepresents14() {
        helpTCF("A.java","public class A {\n int i; //@ represents i = 0;\n}"
                ,"/A.java:2: The target of a represents clause must be a model field",13
                );
    }
    
    public void testRepresents13a() {
        addMockFile("$A/A.jml","public class A {\n //@ ghost int i; represents i = 0;\n}");
        helpTCF("A.java","public class A {\n }"
                ,"/$A/A.jml:2: The target of a represents clause must be a model field",19
                );
    }
    
    public void testRepresents14a() {
        addMockFile("$A/A.jml","public class A {\n //@ represents i = 0;\n}");
        helpTCF("A.java","public class A {\n int i; \n}"
                ,"/$A/A.jml:2: The target of a represents clause must be a model field",6
                );
    }

    /** Check that the target of a static represents clause is in the same class */
    public void testRepresents15() {
        helpTCF("A.java","public class A extends B {\n //@ static represents i = 0;\n} class B { //@ static model int i; \n}"
                ,"/A.java:2: A represents clause must be declared in the same class as the static model field it represents",13
                );
    }
    
    /** Check that the rhs of a static represents clause contains only static fields */
    public void testRepresents16() {
        helpTCF("A.java","public class A {\n static int k; int j; //@ model static int i; static represents i = k;\n}"
                );
    }
    
    /** Check that the rhs of a static represents clause contains only static fields */
    public void testRepresents16a() {
        helpTCF("A.java","public class A {\n static int k; int j; //@ model static int i; static represents i = j;\n}"
                ,"/A.java:2: non-static variable j cannot be referenced from a static context",69
                );
    }
    
    /** Check that warning that <- is deprecated */
    public void testRepresents17() {
        expectedExit = 0;
        helpTCF("A.java","public class A {\n static int j; //@  model static int i; static represents i <- j;\n}"
                ,"/A.java:2: warning: The left arrow is deprecated in represents clauses, use = instead",61
                );
    }
    

    public void testMisc() {
        helpTCF("A.java","public class A {\n //@ ensures ((boolean)\\result);\n int m() { return 0; }}"
                ,"/A.java:2: inconvertible types\n  required: boolean\n  found:    int",24
        );
    }
    
    public void testMisc2() {
        helpTCF("A.java","public class A {\n //@ ensures ((short)\\result) == 0;\n int m() { return 0; }}"
        );
    }
    
    public void testMisc3() {
        helpTCF("A.java","public class A {\n //@ public normal_behavior ensures true; public model boolean m(); \n }"
        );
    }
    
    public void testForall() {
        helpTCF("A.java","public class A {\n //@ forall int i,j; old boolean k=true, m = false; requires i == 0; \n public void m() {}}"
                );
    }
    
    public void testForall2() {
        helpTCF("A.java","public class A {\n //@ forall int i=0,j; old boolean k, m = false; requires i == 0; \n public void m() {}}"
                ,"/A.java:2: A forall method clause declaration must not have initializers",19
                ,"/A.java:2: A old method clause variable must have an initializer",36
        );
    }
    
    public void testForall3() {
        helpTCF("A.java","public class A {\n //@ old int i=true; old boolean m=0; requires i == 0; \n public void m() {}}"
                ,"/A.java:2: incompatible types\n  required: int\n  found:    boolean",16
                ,"/A.java:2: incompatible types\n  required: boolean\n  found:    int",36
                );
    }
    
    public void testForall4() {
        helpTCF("A.java","public class A {\n //@ forall int j; old int k=0; requires i+j<k; \n public void m(int i) {}}"
                );        // OK
    }
    
    public void testForall5() {
        helpTCF("A.java","public class A {\n //@ forall boolean j; old boolean  k=true; requires i+j<k; \n public void m(boolean i) {}}"
                ,"/A.java:2: bad operand types for binary operator '+'\n  first type:  boolean\n  second type: boolean",55
                );        
    }
    
    public void testForall6() {
        helpTCF("A.java","public class A { int i,j,k; \n //@ forall boolean j; old boolean k=true; requires i+j<k; \n public void m(boolean i) {}}"
                ,"/A.java:2: bad operand types for binary operator '+'\n  first type:  boolean\n  second type: boolean",54
                );        
    }
    
    public void testForall7() { // FIXME - this is not the clearest error message - it should refer to the specifications
        helpTCF("A.java","public class A { \n //@ forall int i; old int k=0   ; requires i<k; \n public void m(int i) {}}"
                ,"/A.java:2: i is already defined in m(int)",17
                );        
    }
    
    public void testForall8() { // FIXME - this is not the clearest error message - it should refer to the specifications
        helpTCF("A.java","public class A { \n //@ forall int k; old int k=0   ; requires i<k; \n public void m(int i) {}}"
                ,"/A.java:2: k is already defined in m(int)",28
                );        
    }
    
    public void testForall9() { // FIXME - this is not the clearest error message - it should refer to the specifications
        helpTCF("A.java","public class A { \n //@ forall int j; old int k=0   ; requires i<k; \n//@{| forall int m; ensures k<m; also ensures k<m; |} \n public void m(int i) {}}"
                ,"/A.java:3: cannot find symbol\n  symbol:   variable m\n  location: class A",49
                );        
    }
    
    public void testForall10() { // FIXME - this is not the clearest error message - it should refer to the specifications
        helpTCF("A.java","public class A { \n //@ forall int j; old int k=0   ; requires i<k; \n//@{| forall int k; ensures k<m; also ensures i==0; |} \n public void m(int i) {}}"
                ,"/A.java:3: k is already defined in m(int)",18
                ,"/A.java:3: cannot find symbol\n  symbol:   variable m\n  location: class A",31
                );        
    }


    public void testSignals() { //OK
        helpTCF("A.java","public class A {\n//@signals(Exception e) true; \n void m(){}}");
    }


    public void testSignals1() {//OK
        helpTCF("A.java","public class A {\n//@signals(Exception) true; \n void m(){}}");
    }


    public void testSignals2() { //Bad type
        helpTCF("A.java","public class A {\n//@signals(Object e) true; \n void m(){}}",
                "/A.java:2: incompatible types\n  required: java.lang.Exception\n  found:    java.lang.Object",12);
    }


    public void testSignals3() { //Bad syntax
        helpTCF("A.java","public class A {\n//@signals true; \n void m(){}}",
                "/A.java:2: Expected a left parenthesis after a signals keyword",12);
    }


    public void testSignals4() { //OK
        helpTCF("A.java","public class A {\n//@signals(RuntimeException ) ; \n void m(){}}"
                );
    }


    public void testSignals5() { //Bad type
        helpTCF("A.java","public class A {\n//@signals(java.io.IOException e) 2; \n void m(){}}",
                "/A.java:2: incompatible types\n  required: boolean\n  found:    int",35);
    }

    public void testSignals6() { //Bad type
        helpTCF("A.java","public class A {\n//@signals(int e) true; \n void m(){}}",
                "/A.java:2: incompatible types\n  required: java.lang.Exception\n  found:    int",12);
    }

    public void testSignals7() { //OK - scoping
        helpTCF("A.java","public class A {\n//@signals(java.io.IOException e) e==null; \n void m(){}}");
    }

    public void testSignalsOnly() { //OK
        helpTCF("A.java","public class A {\n//@signals_only \\nothing;\nvoid m() {}}");
    }

    public void testSignalsOnly1() { //OK
        helpTCF("A.java","public class A {\n//@signals_only RuntimeException;\nvoid m() {}}");
    }

    public void testSignalsOnly2() { //OK
        helpTCF("A.java","public class A {\n//@signals_only RuntimeException,Exception;\nvoid m() {}}");
    }

    public void testSignalsOnly3() {
        helpTCF("A.java","public class A {\n//@signals_only ;\nvoid m() {}}",
                "/A.java:2: Use \\nothing to denote an empty list of exceptions in a signals_only clause",17);
    }

    public void testSignalsOnly4() {
        helpTCF("A.java","public class A {\n//@signals_only RuntimeException java.lang.Exception;\nvoid m() {}}",
                "/A.java:2: Missing comma or otherwise ill-formed type name",34);
    }

    public void testSignalsOnly5() {
        helpTCF("A.java","public class A {\n//@signals_only RuntimeException,;\nvoid m() {}}",
                "/A.java:2: illegal start of type",34);
    }

    public void testSignalsOnly6() {
        helpTCF("A.java","public class A {\n//@signals_only RuntimeException,,RuntimeException;\nvoid m() {}}",
                "/A.java:2: illegal start of type",34);
    }

    public void testSignalsOnly7() {
        helpTCF("A.java","public class A {\n//@signals_only RuntimeException\nvoid m() {}}",
                "/A.java:2: Invalid expression or missing semicolon here",33);
    }

    public void testSignalsOnly8() {
        helpTCF("A.java","public class A {\n//@signals_only RuntimeException[];\nvoid m() {}}",
                "/A.java:2: incompatible types\n  required: java.lang.Exception\n  found:    java.lang.RuntimeException[]",33);
    }

    public void testSignalsOnly9() {
        helpTCF("A.java","public class A {\n//@signals_only int;\nvoid m() {}}",
                "/A.java:2: incompatible types\n  required: java.lang.Exception\n  found:    int",17);
    }

    public void testSignalsOnly10() {
        helpTCF("A.java","public class A {\n//@signals_only Q;\nvoid m() {}}",
                "/A.java:2: cannot find symbol\n  symbol:   class Q\n  location: class A",17);
    }
    
    public void testIn() {
        helpTCF("A.java","public class A {\n //@ model JMLDataGroup k; \n int n; //@ in k; \n}"
                );
    }
    
    public void testIn2() {
        helpTCF("A.java","public class A extends B{\n /*@ spec_public */ protected int k; \n int n; //@ in k, this.k, super.kk; \n} class B { //@ model int kk; \n}"
                ,"/A.java:3: Datagroups in \"in\" and \"maps\" clauses must be model variables",16
                ,"/A.java:3: Datagroups in \"in\" and \"maps\" clauses must be model variables",19
                );
    }
    
    public void testIn3() {
        helpTCF("A.java","public class A {\n /*@ spec_public */ protected int k; \n int n; //@ in m; \n}"
                ,"/A.java:3: cannot find symbol\n  symbol:   variable m\n  location: class A",16
                );
    }
    
    public void testIn4() {
        helpTCF("A.java","public class A {\n //@ model static int m; \n int n; //@ in m; \n}"
                ,"/A.java:3: A non-static variable may not be in a static datagroup",16
                );
    }
    
    public void testMaps() {
        helpTCF("A.java","public class A {\n //@ model JMLDataGroup k; \n A next; //@ maps next.next \\into k; \n}"
        );
    }
    
    public void testMaps2() {
        helpTCF("A.java","public class A {\n //@ model JMLDataGroup k; \n A[] next; //@ maps next[*].next \\into k; \n}"
        );
    }
    
    public void testMaps2b() {
        helpTCF("A.java","public class A {\n //@ model JMLDataGroup k; \n A[] next; //@ maps next[*] \\into k; \n}"
        );
    }
    
    public void testMaps3() {
        helpTCF("A.java","public class A {\n //@ model JMLDataGroup k; \n A[] next; //@ maps next[2 .. 3].next \\into k,k; \n}"
        );
    }
    
    public void testMaps4() {
        helpTCF("A.java","public class A {\n //@ model JMLDataGroup k; \n A[] next; //@ maps next[2].next \\into this.k; \n}"
        );
    }
    
    // TODO - should have some tests checking recovery in maps clauses
    
    public void testInitializer() {
        helpTCF("A.java","public class A {\n //@initializer static_initializer \n}"
        );
    }

    public void testInitializer1() {
        helpTCF("A.java","public class A {\n //@initializer static_initializer initializer static_initializer\n}"
                ,"/A.java:2: Only one initializer specification and one static_initializer specification are allowed",36
                ,"/A.java:2: Only one initializer specification and one static_initializer specification are allowed",48
        );
    }

    /** Tests that specs get associated with the initializer */
    public void testInitializer2() {
        helpTCF("A.java","public class A {\n int i; static int j; //@ ensures i==0; initializer ensures j == 0; static_initializer \n}"
        );
    }

    /** Tests that variable references in a static initializer must be static */
    public void testInitializer3() {
        helpTCF("A.java","public class A {\n int i; static int j; //@ ensures i == 0; static_initializer \n}"
                ,"/A.java:2: non-static variable i cannot be referenced from a static context",35
        );
    }

    public void testInitializer4() {
        addMockFile("$A/A.jml","public class A {\n int i; static int j; //@ ensures i == 0; static_initializer \n}");
        helpTCF("A.java","public class A {\n int i; static int j;  \n}"
                ,"/$A/A.jml:2: non-static variable i cannot be referenced from a static context",35
        );
    }

    public void testInitializer5() {
        addMockFile("$A/A.jml","public class A {\n int i; static int j; static {} \n}");
        helpTCF("A.java","public class A {\n int i; static int j;  \n}"
                ,"/$A/A.jml:2: Initializer blocks are not allowed in specifications",30
        );
    }

    public void testInitializer6() {
        helpTCF("A.java","public class A {\n {} static {} \n}"
        );
    }

    public void testInitializer7() {
        helpTCF("A.java","public class A {\n //@ {} static {} {} static {}\n}"
        );
    }

    /** Tests that specs are parsed with the Java initializer */
    public void testInitializer8() {
        helpTCF("A.java","public class A {\n int i; static int j; //@ ensures i==0; \n {} //@ ensures i==0; \n static {} \n}"
                ,"/A.java:3: non-static variable i cannot be referenced from a static context",17
        );
    }

    public void testInitializer9() {
        helpTCF("A.java","public class A {int i; static int j; \n \n static { i = 0; } \n}"
                ,"/A.java:3: non-static variable i cannot be referenced from a static context",11
        );
    }

    public void testInitializer10() {
        helpTCF("A.java","public class A {int i; static int j; \n //@ ensures i==0; \n static { i = 0; } \n}"
                ,"/A.java:3: non-static variable i cannot be referenced from a static context",11
                ,"/A.java:2: non-static variable i cannot be referenced from a static context",14
        );
    }

    // OK
    public void testMonitorsFor() {
        helpTCF("A.java","public class A {A a; Object i,j,k; //@ monitors_for i = j,a.k,Object.class; \n }"
        );
    }
    
    public void testMonitorsFor1() {
        helpTCF("A.java","public class A {Object i,j,k; int m; //@ monitors_for i = 5; \n }"
                ,"/A.java:1: The expression in a monitors_for clause must have reference type",59
        );
    }
    
    public void testMonitorsFor2() {
        helpTCF("A.java","public class A {Object i,j,k; int m; //@ monitors_for i <- m; \n }"
                ,"/A.java:1: The expression in a monitors_for clause must have reference type",60
        );
    }
    
    // OK - static does not matter
    public void testMonitorsFor3() {
        helpTCF("A.java","public class A {Object i,j; static Object k;  //@ monitors_for k = i,A.k; \n }"
        );
    }
    
    public void testMonitorsFor4() {
        helpTCF("A.java","public class A {Object i,j; static Object k;  //@ monitors_for k = Object; \n }"
                ,"/A.java:1: cannot find symbol\n  symbol:   variable Object\n  location: class A",68
        );
    }
    
    public void testMonitorsFor5() {
        helpTCF("A.java","public class A extends B {Object i,j; static Object k;  //@ monitors_for z = i; \n } class B { public Object z; }"
                ,"/A.java:1: The identifier must be a member of the enclosing class: z is in B instead of A",74
        );
    }
    
    //OK
    public void testReadable() {
        helpTCF("A.java","public class A extends B {Object i,j; static Object k;  //@ readable j if i == null; writable j if i == null; \n } class B { public Object z; }"
        );
    }
    
    public void testReadable1() {
        helpTCF("A.java","public class A extends B {Object i,j; static Object k;  //@ readable k if i == null; writable k if i == null; \n } class B { public Object z; }"
                ,"/A.java:1: non-static variable i cannot be referenced from a static context",75
                ,"/A.java:1: non-static variable i cannot be referenced from a static context",100
        );
    }
    
    public void testReadable2() {
        helpTCF("A.java","public class A extends B {Object i,j; static Object k;  //@ readable z if i == null; writable z if i == null; \n } class B { public Object z; }"
                ,"/A.java:1: The identifier must be a member of the enclosing class: z is in B instead of A",70
                ,"/A.java:1: The identifier must be a member of the enclosing class: z is in B instead of A",95
        );
    }

    //OK
    public void testReadable3() {
        helpTCF("A.java","public class A extends B {Object i,j; static Object k;  //@ readable k if k == null; writable k if k == null; \n } class B { public Object z; }"
        );
    }
    
    //OK
    public void testReadable4() {
        helpTCF("A.java","public class A extends B {Object i,j; static Object k;  //@ readable i if this == null; writable i if this == null; \n } class B { public Object z; }"
        );
    }
    
    public void testReadable5() {
        helpTCF("A.java","public class A extends B {Object i,j; static Object k;  //@ readable k if this == null; writable k if this == null; \n } class B { public Object z; }"
                ,"/A.java:1: non-static variable this cannot be referenced from a static context",75
                ,"/A.java:1: non-static variable this cannot be referenced from a static context",103
        );
    }
    
    //OK
    public void testReadable6() {
        helpTCF("A.java","public class A extends B {Object i,j; static Object k;  //@ readable k if Object.class == null; writable k if Object.class == null; \n } class B { public Object z; }"
        );
    }
    

    
}

