package resanaplugin.costa;

import java.util.LinkedList;
import java.util.List;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Dialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;

import resanaplugin.Method;

public class CostaOptionsDialog extends Dialog {

	public CostaOptionsDialog(Shell sh) {
		super(sh);
	}

	public CostaOptions createWindow(List<Method> methods) {
		final CostaOptions result = new CostaOptions();
		Shell parent = getParent();
		final Shell shell = new Shell(parent, SWT.TITLE | SWT.BORDER
				| SWT.APPLICATION_MODAL);
		shell.setText("Costa options");
		shell.setLayout(new GridLayout(2, true));
		shell.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

		Group methodsGroup = new Group(shell, SWT.NONE);
		methodsGroup.setLayout(new GridLayout(1, false));
		GridData data = new GridData(SWT.FILL, SWT.FILL, false, false);
		data.verticalSpan = 2;
		methodsGroup.setLayoutData(data);
		methodsGroup.setText("Methods");
		final List<Button> methodButtons = new LinkedList<Button>();
		for (Method method : methods) {
			Button newButton = new Button(methodsGroup, SWT.CHECK);
			newButton.setText(method.getMethodSignature(true));
			newButton.setData(method);
			newButton.setSelection(true);
			methodButtons.add(newButton);
		}

		Composite selectButtons = new Composite(methodsGroup, SWT.NONE);
		selectButtons.setLayout(new GridLayout(2, true));
		selectButtons.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false,
				false));

		final Button selectAll = new Button(selectButtons, SWT.PUSH);
		selectAll.setText("Select All");
		selectAll.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				for (Button methodButton : methodButtons) {
					methodButton.setSelection(true);
				}
			}
		});

		final Button deselectAll = new Button(selectButtons, SWT.PUSH);
		deselectAll.setText("Deselect All");
		deselectAll.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				for (Button methodButton : methodButtons) {
					methodButton.setSelection(false);
				}
			}
		});

		Group optionsGroup = new Group(shell, SWT.NONE);
		optionsGroup.setLayout(new GridLayout(2, false));
		optionsGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false,
				false));
		optionsGroup.setText("Options");

		final Label memoryUnitLabel = new Label(optionsGroup, SWT.NONE);
		memoryUnitLabel.setText("Memory Unit: ");

		final Combo memoryUnitCombo = new Combo(optionsGroup, SWT.DROP_DOWN
				| SWT.READ_ONLY);
		memoryUnitCombo.setItems(new String[] { "heap", "stack" });
		memoryUnitCombo.setText("heap");

		Group boundsGroup = new Group(shell, SWT.NONE);
		boundsGroup.setLayout(new GridLayout(1, false));
		boundsGroup
				.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		boundsGroup.setText("Heap bound options");

		final Button costaBound = new Button(boundsGroup, SWT.CHECK);
		costaBound.setText("Costa heap bound");
		costaBound.setSelection(false);

		final Button simpleBound = new Button(boundsGroup, SWT.CHECK);
		simpleBound.setText("Simple heap bound");
		simpleBound.setSelection(true);

		Composite concreteGroup = new Composite(boundsGroup, SWT.NONE);
		concreteGroup.setLayout(new GridLayout(3, false));
		concreteGroup
				.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		
		final Button jamaicaBound = new Button(concreteGroup, SWT.CHECK);
		jamaicaBound.setText("Concrete heap bound");
		jamaicaBound.setSelection(false);
		
		final Label vmLabel = new Label(concreteGroup, SWT.NONE);
		vmLabel.setText("VM: ");

		final Combo vmCombo = new Combo(concreteGroup, SWT.DROP_DOWN
				| SWT.READ_ONLY);
		vmCombo.setItems(new String[] { "jamaica", "host" });
		vmCombo.setText("jamaica");
		vmCombo.setEnabled(memoryUnitCombo.getText() == "heap" && jamaicaBound.getSelection());
		
		jamaicaBound.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				vmCombo.setEnabled(memoryUnitCombo.getText() == "heap" && jamaicaBound.getSelection());
			}
		});

		memoryUnitCombo.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				if (memoryUnitCombo.getText() == "stack") {
					costaBound.setEnabled(false);
					simpleBound.setEnabled(false);
					jamaicaBound.setEnabled(false);
				} else {
					costaBound.setEnabled(true);
					simpleBound.setEnabled(true);
					jamaicaBound.setEnabled(true);
				}
			}
		});

		new Label(shell, SWT.NONE);

		Composite rightButtons = new Composite(shell, SWT.FILL);
		rightButtons
				.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		rightButtons.setLayout(new FillLayout());

		final Button cancel = new Button(rightButtons, SWT.PUSH);
		cancel.setText("cancel");
		cancel.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				shell.dispose();
			}
		});

		final Button doAnalysis = new Button(rightButtons, SWT.PUSH);
		doAnalysis.setText("Do analysis");
		doAnalysis.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				for (Button methodButton : methodButtons) {
					if (methodButton.getSelection()) {
						result.methods.add((Method) methodButton.getData());
					}
				}
				if (memoryUnitCombo.getText() == "stack")
					result.measuredBy = true;
				else
					result.measuredBy = false;
				result.vm = vmCombo.getText();
				if (costaBound.isEnabled())
					result.costaBound = costaBound.getSelection();
				else
					result.costaBound = false;
				if (simpleBound.isEnabled())
					result.simpleBound = simpleBound.getSelection();
				else
					result.simpleBound = false;
				if (jamaicaBound.isEnabled())
					result.concreteBound = jamaicaBound.getSelection();
				else
					result.concreteBound = false;
				shell.dispose();
			}
		});

		shell.pack();
		shell.open();

		Display display = parent.getDisplay();
		while (!shell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}

		return result;
	}
}
