package resanaplugin.loopbounds;

import java.io.StringWriter;
import loopbounds.LoopBoundsAnalyzer;
import loopbounds.parsetree.*;

import org.eclipse.jface.action.IAction;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorActionDelegate;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.jdt.internal.ui.javaeditor.JavaEditor;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.ui.console.*;

import resanaplugin.ConsoleMessager;
import resanaplugin.ProjectInformation;

/**
 * 
 * @author Nick Gubbels
 * @author Liviu Varga
 * @author Joel Haasnoot
 * @author Rody Kersten
 * 
 */

@SuppressWarnings("restriction")
public class EditorActionDelegate1 implements IEditorActionDelegate {

	private ProjectInformation projectInf = new ProjectInformation();
	private LoopBoundsAnalyzer lba;

	@Override
	public void selectionChanged(IAction action, ISelection selection) {
		projectInf.changeSelection(selection);
	}

	@Override
	public void setActiveEditor(IAction action, IEditorPart targetEditor) {
		if (targetEditor instanceof JavaEditor) {
			projectInf.setActiveEditor((JavaEditor) targetEditor);
		}
	}

	@Override
	/**
	 * Main plug-in functionality, this is run at every push of the button of the plug-in
	 * It looks at the position of the cursor in the editor window and looks for a loop
	 * which starts at current or previous index
	 * 
	 * @param	action
	 */
	public void run(IAction action) {
		String s;
		int cc = 0, mm = 0, ll = 0;
		Options options;

		MessageConsole myConsole = ConsoleMessager.findConsole("RESANA");
		MessageConsoleStream out = myConsole.newMessageStream();
		Shell sh = PlatformUI.getWorkbench().getActiveWorkbenchWindow()
				.getShell();
		if (projectInf.save(sh)) {
			if (projectInf.errors()) {
				ConsoleMessager.compilationErrorDialog(sh);
			} else {
				InputOptionsDialog in = new InputOptionsDialog(sh);
				StringWriter diagnostic = new StringWriter();

				ISourceViewer viewer = projectInf.getViewer();
				IDocument doc = viewer.getDocument();
				Point p = viewer.getSelectedRange();

				try {
					lba = new LoopBoundsAnalyzer(diagnostic);
					lba.setClasspath(ProjectInformation
							.getClassPathArray(projectInf.getProject()));
					ParseTreeWrapper input = new ParseTreeWrapper(diagnostic);
					input.parseSource(doc.get());

					if (action.getId().endsWith("analyseAll")) {
						options = in.open();

						// Check if cancelled
						if (!options.ok)
							return;

						s = lba.analyseAllLoops(input, options.degree,
								options.step, options.branchSplitting,
								options.correctJML, options.overwrite,
								options.nodeStartPoint, options.nodeMinVal,
								options.nodeMaxVal, options.refactoring);

						out.println(diagnostic.toString());
						writeSourceBack(s);

					} else if (action.getId().endsWith("analyse")) {

						int Position = p.x;

						// find nearest loop before 'Position'
						searchloop: for (int c = 0; c < input.getNumClasses(); c++) {
							for (int m = 0; m < input.getNumMethods(c); m++) {
								for (int l = 0; l < input.getNumLoops(c, m); l++) {

									if (Position >= input.getLoopPosition(c, m,
											l)) {
										cc = c;
										mm = m;
										ll = l;
									} // looks for the nearest loop with the
										// current
										// index (which starts before at exactly
										// at
										// the position)
									else
										break searchloop;
								}
							}
						}

						// Open options dialog
						options = in.open();

						// Check if cancelled
						if (!options.ok)
							return;

						s = lba.analyseLoop(input, cc, mm, ll, options.degree,
								options.step, options.branchSplitting,
								options.correctJML, options.overwrite,
								options.nodeStartPoint, options.nodeMinVal,
								options.nodeMaxVal, options.refactoring);

						out.println(diagnostic.toString());
						writeSourceBack(s);
					} else {
						System.err.println("[ResAna] Unknown action ID");
						out.println("[ResAna] Unknown action ID");
						out.println(diagnostic.toString());
						return;
					}
				} catch (java.io.IOException e) {
					out.println(diagnostic.toString());
					e.printStackTrace(new java.io.PrintStream(out));
					System.exit(1);
				} catch (UndeclaredVariableException e) {
					// TODO Auto-generated catch block
					out.println(diagnostic.toString());
					e.printStackTrace(new java.io.PrintStream(out));
				} catch (UnsupportedTypeException e) {
					// TODO Auto-generated catch block
					out.println(diagnostic.toString());
					e.printStackTrace(new java.io.PrintStream(out));
				} catch (Exception e) {
					// TODO Auto-generated catch block
					out.println(diagnostic.toString());
					e.printStackTrace(new java.io.PrintStream(out));
				}
				IWorkbenchPage page = PlatformUI.getWorkbench()
						.getActiveWorkbenchWindow().getActivePage();// obtain
																	// the
																	// active
																	// page
				String id = IConsoleConstants.ID_CONSOLE_VIEW;
				IConsoleView view = null;
				try {
					view = (IConsoleView) page.showView(id);
				} catch (PartInitException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
					out.println(e.toString());
				}
				view.display(myConsole);
			}
		}
	}

	private void writeSourceBack(String s) {
		ISourceViewer viewer = projectInf.getViewer();
		IDocument doc = viewer.getDocument();
		doc.set(s);
	}

}
